;;; Macros for the Scheme 84 compiler in Franz Lisp.

(eval-when (eval compile load) (setsyntax 35 'vcharacter)) ;#  
; added eval Thu Mar  5 11:22:36 1987 mw

;(declare
;   (localf cdr-assq mk-inst emit lookup try-rib-m try-n emit-fake-lambda
;      a-args macro-expand e-exp e-constant e-macro
;      e-if
;      e-application
;      e-stmnts t-exp t-constant t-identifier t-macro t-lambda
;      t-if t-set 
;      t-application t-stmnts f-exp f-lambda p-exp))

(declare
   (special ??? **location** **fnv** core-names table-*at table-*gs
      application-declared-semantics scheme-id-declared-semantics
      **comp-env** table-*gr host-macs))

(setq **comp-env** nil)

(def if
   (macro (l)
	  (cond ((eq (length l) 3) `(cond (,(cadr l) ,(caddr l))))
		((eq (length l) 4) `(cond (,(cadr l) ,(caddr l))
					  (t ,(cadddr l)))))))

(def tag-vars-length
   (lambda (l)
      (cond
	 ((null l) (list '*la nil 0))
	 ((atom l) (list '*rla (list l) 0))
	 (t (let ((x (imlast l)))
	       (cond
		  ((null (cdr x)) (list '*la l (length l)))
		  ((atom (cdr x))
		   (list '*rla 
		      (cons (cdr x) (imrdc l)) 
		      (lengths l)))))))))

(def imrdc
   (lambda (l)
      (cond
	 ((atom l) nil)
	 (t (cons (car l) (imrdc (cdr l)))))))

(def imlast
   (lambda (l)
      (cond
	 ((atom (cdr l)) l)
	 (t (imlast (cdr l))))))

(def lengths
   (lambda (l)
      (cond
	 ((atom l) 0)
	 (t (add1 (lengths (cdr l)))))))


(def purcopy*
   (lambda (s)
      (cond
	 ((atom s)  s)
	 ((purep s) s)
	 (t (let ((x (purcopy (cons nil nil))))
	       (rplaca x (purcopy* (car s)))
	       (rplacd x (purcopy* (cdr s))))))))

(def flatten-comp-env
   (lambda (ce)
      (apply (function append) ce)))

;;; define macro hack to make Lisp look like Scheme.
;;; This macro is only good for defining functions!!!

(def define 
   (macro (f)
	  (let ((name (cadr f))
		(form (caddr f)))
	     (cond  ((atom form)
		     `(eval-when (eval load)
			 (putd ',name (getd ',form))))
		    ((memq (car form) '(lambda macro))
		     `(def ,name ,form))
		    (t (print (list "can't define " name form)))))))
; added eval in eval-when Thu Mar  5 11:35:33 1987 mw
	   
(def cdr-assq
   (lambda (a l)
      (cond ((null l) nil)
	    ((eq (cdar l) a) (car l))
	    (t (cdr-assq a (cdr l))))))

(setq table-*at (purcopy* `((*at . 1))))
(setq table-*gs (purcopy* `((*gs . scheme-directory))))
(setq table-*gr (purcopy* `((*gr . scheme-directory))))

(declare (special columns))
(def make-table
   (lambda (rows columns)
      (cons (cons rows columns)
	    (reverse (mapnum 
			(function (lambda (row) (make-col row columns)))
			rows)))))
(declare (special row))

(def make-col
   (lambda (row columns)
      (reverse (mapnum 
		  (function (lambda (col) (make-entry row col)))
		  columns))))

(def make-entry
   (lambda (i j)
      (let ((box (cons i j)))
	   (cons (cons '*lr box) (cons '*ls box)))))

(def mapnum
   (lambda (f i)
      (cond ((minusp i) nil)
	    (t (cons (apply f (list i)) (mapnum f (sub1 i)))))))

(declare (special lexical-table))

(setq lexical-table (purcopy* (make-table 4 4)))
      


(def local-ref-table-lookup
   (lambda (opcode args)
      (let ((rows (caar lexical-table))
	    (columns (cdar lexical-table))
	    (array (cdr lexical-table))
	    (m (car args))
	    (n (cdr args)))
	   (cond ((and (or (lessp m rows) (= m rows))
		       (or (lessp n columns) (= n columns)))
		  (cond ((eq opcode '*lr) (car (nthelem (1+ n)
						  (nthelem (1+ m) array))))
			((eq opcode '*ls) (cdr (nthelem (1+ n)
						  (nthelem (1+ m) array))))))
		 (t (cons opcode args))))))

(def table-lookup
   (lambda (table opcode args)
      (let ((pair (cdr-assq args table)))
	   (cond ((null pair) (let ((newpair (cons opcode args)))
				   (nconc table (list newpair))
				   newpair))
		 (t pair)))))

(def access-system-function
   (macro (l) 
      `(let ([pair 
	       (getl ,(cadr l)
		  '(constant-system-function system-function))])
	 (cond
	    [pair (cadr pair)]
	    [t nil]))))

(def access-scheme-primitive
   (macro (l) 
      `(let ([pair 
	       (getl ,(cadr l)
		  '(constant-primitive scheme-primitive))])
	 (cond
	    [pair (cadr pair)]
	    [t nil]))))

(define one (macro (l) `(car ,(cadr l))))

(define two (macro (l) `(cadr ,(cadr l))))

(define three (macro (l) `(caddr ,(cadr l))))

(define four (macro (l) `(cadddr ,(cadr l))))

(define la-body (macro (l) `(cdddr ,(cadr l))))

(def extend (macro (l) (list 'cons (three l) (two l))))

(define &arity (macro (l) `(car ,(cadr l))))

(define &class (macro (l) `(caddr ,(cadr l))))

(define &name (macro (l) `(cdddr ,(cadr l))))


;;; Expression type predicates.  One fore each type of expression

(define quoted? 
   (macro (l) `(let ((temp ,(cadr l)))
		    (and (dtpr temp)
			 (eq (car temp) 'quote)
			 t))))
(def constant?
   (lambda (x)
      (if (atom x)
	  (or (null x) 
	      (eq x t) 
	      (numberp x) 
	      (stringp x))
	  (or (proc? x) (quoted? x)))))

(define identifier? (macro (l) `(atom ,(cadr l))))

(def lambda? (macro (l) (list 'eq (cons 'car (cdr l)) ''#!lambda)))

(def if? (macro (l) (list 'eq (cons 'car (cdr l)) ''#!if)))

(def set? (macro (l) (list 'eq (cons 'car (cdr l)) ''#!set!)))

(def scheme-id? (macro (l) (list 'eq (cons 'car (cdr l)) ''#!scheme-id)))

(def app? (macro (l) (list 'eq (cons 'car (cdr l)) ''#!application)))

(def host-mac-exp?
   (macro (l) `(and (atom (car ,(cadr l)))
		    (memq  (car ,(cadr l)) host-macs))))

(def beta-transform?
   (macro (l)
      `(let ((exp ,(cadr l)))
	    (if (atom exp) nil
		(if (atom (car exp))
		    (get (car exp) 'beta-transform)
		    (eq (car (car exp)) '&transform))))))
		
(def application? (macro (l) 't))

;;;


;;; code sequence construction


(def inst 
   (macro (l) 
	  (let ((num-args (length l)))
	       (cond ((= num-args 1) 
		      (warn "not enough args to inst" l))
		     ((= num-args 2) `(mk-inst ,(cadr l) nil))
		     ((= num-args 3) `(mk-inst . ,(cdr l)))
		     (t (warn "too many args to inst " l))))))


(def mk-inst
   (lambda (opcode args)
      (cond ((eq opcode '*at) 
	     (cond
		((zerop args) '(*ti . 0))
		(t (table-lookup table-*at opcode args))))
	    ((eq opcode '*lr) 
	     (local-ref-table-lookup opcode args))
	    ((eq opcode '*ls) 
	     (local-ref-table-lookup opcode args))
	    ((eq opcode '*gr) 
	     (table-lookup table-*gr opcode args))
	    ((eq opcode '*gs) 
	     (table-lookup table-*gs opcode args))
	    ((eq opcode '*pu) '(*pu))
	    ((eq opcode '*re) '(*re))  
	    (t (cons opcode args)))))

(def emit
      (lambda (inst code)
      `(,inst . ,code)))



;;; environment access

(def lookup
    (lambda (i env)
	(try-rib-m i env 0)))

(def try-rib-m
   (lambda (i e m)
      (if (null e)
	  nil
	  (let ((place-in-rib 
		   (cond ((not (atom (one e)))
			  (try-n i (one e) 0))
			 ((eq i (one e)) 0)
			 (t nil))))
	       (if place-in-rib
		   (cons m place-in-rib)
		   (try-rib-m i (cdr e) (add1 m)))))))

(def try-n
   (lambda (i r n)
      (cond 
	 ((null r) nil)
	 ((eq i (one r)) n)
	 (t (try-n i (cdr r) (add1 n))))))



(def emit-fake-lambda
   (lambda (p primitive)
      (let ((numargs (car p)))
	 ;;; this 'car' is a KLUDGE, because I am too lazy to restructure
	 ;;; e-identifier.   --brooks
	 (car (cond
		 ((= numargs 0)
		  (e-exp `(#!lambda ()
			     (#!application (,primitive))) nil nil))
		 ((= numargs 1)
		  (e-exp `(#!lambda (x)
			     (#!application 
				(,primitive (#!scheme-id x))))
		     nil nil))
		 ((= numargs 2)
		  (e-exp `(#!lambda (x y)
			     (#!application
				(,primitive (#!scheme-id x) (#!scheme-id y))))
		     nil nil))
		 ((= numargs 3)
		  (e-exp `(#!lambda (x y z)
			     (#!application
				(,primitive
				   (#!scheme-id x)
				   (#!scheme-id y)
				   (#!scheme-id z))))
		     nil nil))
		 (t (raise (list 'SE%comp 0 t 'compile:
			      '|no primitive has this many arguments| numargs))))))))


(def primop
   (lambda (exp env)
      (if (and (atom exp)
	       (not (numberp exp))
	       (not (lookup exp env)))
	  (access-scheme-primitive exp)
	  nil)))



;;; compilation routines


(def compile                  ;; user interface
   (lambda (exp)
      (e-exp exp nil (emit (inst '*re nil) nil))))

;;; argument compiler

(def a-args 
   (lambda (args env code) 
      (cond ((null args) code)
	    (t (a-args (cdr args) 
		  env
		  (t-exp (car args)
			 env
			 (emit (inst '*pu)
			       code)))))))

(def macro-expand
   (lambda (exp env)
      ;;; compile is a primitive so vsm regs they must be bound
      ;;; before macro expansion so that they are not "overwritten"
      ;;; upon completion of macro expansion.	This is especially 
      ;;; true of **fnv**.  If it is not rebound, macro expansion 
      ;;; termination will leave nil in **fnv**.  This will cause 
      ;;; an error on the next fluid lookup, which will in turn 
      ;;; reset **fnv** to something reasonable. 
      ;;; -- gsb 12/27/83
      (setq fal nil)
      (let ((ans
	       (newnames
		  (lexpand2
		     (copy-no-constant-no-quote
			(let ((**fnv** **fnv**))
			   (vsm '((*aa))
				 '((((*ha))) nil)
				 nil
				 (list (tag-frees (copy-no-constant exp) nil))
				 (if (atom (car exp))
				     (caddr (get (car exp) 'beta-transform))
				     (caddar exp))))))
		  nil)))
	 (setq fal nil)
	 ans)))
		     




;;; expression compiler for tail-recursive expressions

(def e-exp 
   (lambda (exp env code)
      (cond ((constant? exp) (standard-e-constant exp env code))
	    ((identifier? exp) (effected-e-identifier exp env code))
	    ((scheme-id? exp) (standard-e-identifier (cadr exp) env code))
	    ((beta-transform? exp) (e-macro exp env code))
	    ((host-mac-exp? exp) (e-exp (host-mac-dispatch exp) env code))
	    ((lambda? exp) (e-lambda exp env code))
	    ((if? exp) (e-if exp env code))
	    ((set? exp) (e-set exp env code))
	    ((app? exp) (standard-e-application (cadr exp) env code))
	    ((application? exp) (effected-e-application exp env code))
	    (t ???))))

(def standard-e-constant        ;;; constants are always tidy
   (lambda (exp env code)
      (emit (cond ((or (atom exp) 
		       (proc? exp))
		   (inst '*co exp))
		  ((quoted? exp)
		   (inst '*co (two exp)))
		  (t (raise (list 'SE%comp 0 t 'compile:
			       '|Bad constant| exp))))
	    code)))

(def run-semantics
   (lambda (exp env semantic-fn)
      (setq **comp-env** (cons (flatten-comp-env env) **comp-env**))
      (let ([**fnv** **fnv**])
	       (let ([ans (vsm '((*aa))
		          '((((*ha))) nil)
		          nil
		          (list exp)
		          semantic-fn)])
		  (setq **comp-env** (cdr **comp-env**))
		  ans))))

; primitives and system functions as global references need special handling
; they cannot be rebound at the top level

(def effected-e-identifier
   (lambda (exp env code)
      (if  scheme-id-declared-semantics
	 (e-exp
	    (run-semantics exp env scheme-id-declared-semantics)
	    env
	    code)
	 (standard-e-identifier exp env code))))

(def standard-e-identifier
   (lambda (exp env code)
    (let ((m-n (lookup exp env))) ;lookup returns dotted-pair
      (emit (if m-n
	       (inst '*lr m-n)
	       (let ([pair (getl exp '(constant-primitive
					 scheme-constant
					 constant-system-function
					 system-function
					 ))])
		  (cond
		     [pair 
			(cond
			   [(eq (cadr pair) 'unassigned-constant)
			    (inst '*co `(&unassigned-constant . ,exp))]
			   [(eq (car pair) 'constant-primitive)
			    (emit-fake-lambda (cadr pair) exp)]
			   [(eq (car pair) 'scheme-primitive)
			    (emit-fake-lambda (cadr pair) exp)]
			   [(eq (car pair) 'system-function)
			    (inst '*co `(&sys . ,(cadr pair)))]
			   [(eq (car pair) 'constant-system-function)
			    (inst '*co 
			       `(&sys . ,(cadr pair)))]
			   [t (inst '*co (cadr pair))])]
		     [t (inst '*gr exp)])))
	    code))))
	
(def e-lambda
   (lambda (exp env code)
      (let ((tag$vars$length (tag-vars-length (two exp)))
	    (body (cddr exp)))
	 (let ((vars (cadr tag$vars$length)))	 
	    (emit
	       (inst (car tag$vars$length)
		     `(,(caddr tag$vars$length) ,vars 
			,@(e-stmnts
			     body
			     (if (null vars)  ;; thunks don't extend 
				 env		;; the environment
				 (extend env vars))
			     (emit (inst '*re) nil))))
	       code)))))
	      
(def e-if 
   (lambda (exp env code)
      (if (not (or (= (length exp) 3)
		   (= (length exp) 4)))
	  (raise (list 'SE%comp 0 t 'compile:
		    '|wrong number of forms to if:| exp)))
      (t-exp (two exp)
	     env
             (emit (inst '*if 
		      `(,(e-exp (three exp) 
				env 
				code)
			 .,(e-exp (four exp)
				  env
				  code)))
		   nil))))

(def e-set
   (lambda (exp env code)
      (if (not (= (length exp) 3))
	  (raise (list 'SE%comp 0 t 'compile:
		    '|wrong number of forms to set!:| exp)))
      (t-exp (three exp)
	     env
	     (emit (let ((m-n (lookup (two exp) 
				      env))) ;lookup returns `(m . n)
			(cond (m-n (inst '*ls m-n))
			      (t (inst '*gs (two exp)))))		   
		   code))))

(def e-stmnts
   (lambda (stmnts env code)
      (cond ((null (cdr stmnts))
	     (e-exp (car stmnts)
		    env
		    code))
	    (t (t-exp (car stmnts)
		      env
		      (e-stmnts (cdr stmnts) 
			 env
			 code))))))
 
(def standard-e-application
   (lambda (exp env code)
      (let* ((args (cdr exp))
	     (len-args (length args))
	     (fcn (beta-expand (car exp)))
	     (prim (primop fcn env)))
	 (cond
	    ((and prim
		(not (baselocation fcn)))
	     (if (not (= (&arity prim) len-args))
		 (raise (list 'SE%comp 0 t 'compile:
			   '|Wrong number of args to primitive:| exp))
		 nil)
	     (p-exp prim args env code))
	    ((and (lambda? fcn)
		  (let ([tag$vars$length (tag-vars-length (cadr fcn))])
		     (or (and (eq (car tag$vars$length) '*la)
			      (not (= (caddr tag$vars$length) len-args)))
			 (and (eq (car tag$vars$length) '*rla)
			      (lessp len-args (caddr tag$vars$length))))))
	     (raise (list 'SE%comp 0 t 'compile:
		       '|Wrong number of actual parameters:| exp)))
	    (t (a-args args
		  env
		  (f-exp fcn
		     env
		     (emit (inst '*at len-args)
			   nil))))))))

(def effected-e-application
   (lambda (exp env code)
      (if application-declared-semantics
	 (e-exp
	    (run-semantics exp env application-declared-semantics)
	    env
	    code)
	 (standard-e-application exp env code))))

(def e-macro
   (lambda (exp env code)
      (e-exp (macro-expand exp env)
	     env
	     code)))		    	    



;;; expression compiler for non-tail-recursive  expressions
;;; (ie. expressions which must be made tidy.)

(def t-exp 
   (lambda (exp env code)
      (cond ((constant? exp) (standard-t-constant exp env code))
	    ((identifier? exp) (effected-t-identifier exp env code))
	    ((scheme-id? exp) (standard-t-identifier (cadr exp) env code))
	    ((beta-transform? exp) (t-macro exp env code))
	    ((host-mac-exp? exp) (t-exp (host-mac-dispatch exp) env code))
	    ((lambda? exp) (t-lambda exp env code))
	    ((if? exp) (t-if exp env code))
	    ((set? exp) (t-set exp env code))
	    ((app? exp) (standard-t-application (cadr exp) env code))
	    ((application? exp) (effected-t-application exp env code))
	    (t ???))))

(define standard-t-constant standard-e-constant)

(define standard-t-identifier standard-e-identifier)

(def effected-t-identifier
   (lambda (exp env code)
	 (if  scheme-id-declared-semantics
	     (t-exp
		(run-semantics exp env scheme-id-declared-semantics)
		env
		code)
	     (standard-t-identifier exp env code))))

(define t-lambda e-lambda)

(def t-if 
   (lambda (exp env code)
      (if (not (or (= (length exp) 3)
		   (= (length exp) 4)))
	  (raise (list 'SE%comp 0 t 'compile:
		    '|wrong number of forms to if:| exp)))
      (t-exp (two exp)
	     env
             (emit (inst '*if 
		      `(,(t-exp (three exp) 
				env 
				code)
			 .,(t-exp (four exp)
				  env
				  code)))
		   nil))))

(define t-set e-set)

(def t-stmnts 
   (lambda (stmnts env code)
      (cond ((null (cdr stmnts))    ; Just like e-stmnts but does tidy
	     (t-exp (car stmnts)    ; compile on last form.
		    env
		    code))
	    (t (t-exp (car stmnts) 
		      env
		      (t-stmnts (cdr stmnts)
			 env
			 code))))))

(def standard-t-application
   (lambda (exp env code)
      (let* ((args (cdr exp))
	     (len-args (length args))
	     (fcn (beta-expand (car exp)))
	     (prim (primop fcn env)))
	 (cond
	    ((and prim (not (baselocation fcn)))
	     (if (not (= (&arity prim) len-args))
		 (raise (list 'SE%comp 0 t 'compile:
			   '|Wrong number of args to primitive:| exp))
		 nil)
	     (p-exp prim args env code))
	    ((and (lambda? fcn)
		  (let ([tag$vars$length (tag-vars-length (cadr fcn))])
		     (or (and (eq (car tag$vars$length) '*la)
			      (not (= (caddr tag$vars$length) len-args)))
			 (and (eq (car tag$vars$length) '*rla)
			      (lessp len-args (caddr tag$vars$length))))))
	     (raise (list 'SE%comp 0 t 'compile:
		       '|Wrong number of actual parameters:| exp)))
	    (t (emit (inst '*sa code)
		     (a-args args
			env
			(f-exp fcn
			   env
			   (emit (inst '*at len-args)
				 nil)))))))))

(def effected-t-application
   (lambda (exp env code)
      (if application-declared-semantics
	 (t-exp
	    (run-semantics exp env application-declared-semantics)
	    env
	    code)
	 (standard-t-application exp env code))))

(def t-macro
   (lambda (exp env code)
      (t-exp (macro-expand exp env)
	     env
	     code)))


;;; Compiler for expression in functional postion.   This is just like
;;; t-exp, except that we special case lambda (= exp) expressions so that 
;;; a *le optimization can be done.  


;;; Note: If exp is an identifier, then it is guarenteed not to be a 
;;; primitive as they are special-cased in e-application and t-application.

(def f-exp 
   (lambda (exp env code)
      (cond ((constant? exp) (standard-t-constant exp env code))
	    ((identifier? exp) (effected-t-identifier exp env code))
	    ((scheme-id? exp) (standard-t-identifier (cadr exp) env code))
	    ((beta-transform? exp) (t-macro exp env code))
	    ((host-mac-exp? exp) (f-exp (host-mac-dispatch exp) env code))
	    ((lambda? exp) (f-lambda exp env code))  ;; the 
	    ((if? exp) (t-if exp env code))
	    ((set? exp) (t-set exp env code))
	    ((app? exp) (standard-t-application (cadr exp) env code))
	    ((application? exp) (effected-t-application exp env code))
	    (t ???))))

(def f-lambda 
   (lambda (exp env code)
      (let ((tag$vars$length (tag-vars-length (cadr exp))))
	 (let ((vars (cadr tag$vars$length)))
	    (if (null vars)
		(e-stmnts
		   (cddr exp)
		   env
		   (emit (inst '*re) nil))
		(emit (inst (if (eq '*la (car tag$vars$length))
				'*le
				'*rle)
			     vars)
		      (e-stmnts
			 (cddr exp)
			 (extend env vars)
			 (emit (inst '*re) nil))))))))

;;; primitive expression compiler

(def p-exp 
   (lambda (prim args env code)
      (a-args (cdr args)
	 env 
	 (t-exp (car args)
		env
		(emit (cdr prim)
		      code)))))
   
