; standard macros and functions
; 
; this file is designed to be loaded into a bare Scheme
; thus all macros and functions must be defined before they are used

; all the other hand S-coded definitions are at the head of top.l and sys.s

(#!set! list (#!lambda x x))
(declare-constant 'list)

(add-to-syntax-table '(define-constant)
   (lambda (l)
      `(begin
	  (#!set! ,(cadr l) ,(caddr l))
	  (declare-constant ',(cadr l)))))

(define-constant quit '(&cont (((*ha)) nil)))

(define-constant any (lisp-eval 'any))

(define-constant true t)

(define-constant else t)

(define-constant false nil)

(define-constant eof (lisp-eval '(eof)))

(define-constant nat-log-base (exp 1))

(define-constant pi (* 2 (arcsin 1)))

(define-constant >>> 'a_prompt)

(define-constant mkmackeywords 
   (list '... '*pattern* 'with 'withrec
      'by-cases 'on-num-terms 'on-own-cases
      '*generated-symbol* 'weave
      'syntax 'declare-syntax 'declare-syntax/code 'extend-syntax
      'syntactic-transform-function
      'mkmac))



(define-constant thaw (#!lambda (th) (th)))

(define-constant mapcar
   (#!lambda (f l)
      (((#!lambda (mapcar)
	   (#!set! mapcar (#!lambda (l)
			   (#!if l
			      (cons (f (car l)) (mapcar (cdr l)))))))
	any)
       l)))

(define-constant mapc
   (#!lambda (f l)
      (((#!lambda (mapc)
	   (#!set! mapc (#!lambda (l)
			 (#!if l (begin
				  (f (car l))
				  (mapc (cdr l)))))))
	any)
       l)))

(define-constant writeln 
   (#!lambda x
      ((#!lambda (x)
	  (((#!lambda (*00000)
	       (#!set! *00000 (#!lambda nil
			       (#!if x
				   (begin (print (car x))
					  (#!set! x (cdr x))
					  (*00000))
				   (newline)))))
	    any)))
       x)))



(add-to-syntax-table (quote (case else))
   (#!lambda (*00000)
      ((#!lambda (*00000)
	  (let ([tag (car *00000)]
		[pairs (cdr *00000)])
	     (list
		(list (quote #!lambda)
		      (quote (*00000))
		      (iterate loop ([p pairs])
			 (#!if p
			    (#!if (and (null? (cdr p))
				     (eq? (caar p) (quote else)))
				(cons
				   (quote begin)
				   (append (cdar p) (quote nil)))
				(list
				   (quote #!if)
				   (list
				      (#!if (atom? (caar p))
					  (quote eqv?)
					  (quote memv))
				      (quote *00000)
				      (list
					 (quote quote)
					 (caar p)))
				   (cons (quote begin)
					 (append (cdar p) (quote nil)))
				   (loop (cdr p))))
			    (cons
			       (quote begin)
			       (cons
				  (quote (print (quote |[case |)))
				  (cons
				     (list (quote print)
					   (list (quote quote) tag))
				     (cons
					(cons (quote #!if)
					      (cons
						 (list (quote quote) pairs)
						 (quote ((newline)))))
					(cons
					   (list
					      (quote mapc)
					      (quote
						 (#!lambda (x)
						    (writeln
						       (quote |   |)
						       x)))
					      (list (quote quote) pairs))
					   (quote
					      ((writeln (quote |]|))
					       (writeln
						  (quote
						     |[case: unmatched tag: |)
						  *00000
						  (quote |]|))))))))))))
		tag)))
       (cdr *00000))))

; transcript, compile-file, and edit go in patches.s
; they are usually autoloaded, so their initial definition involves
;     a file pointer.  This file contains no file pointers.

; don't delete this innocent looking piece of code
(define-constant apply (#!lambda (f l) (primitive-apply f l)))
(define-constant let-fluid1 (#!lambda (x y z)
			       (primitive-let-fluid1 x y z)))

(define-constant engine-return
   ((engine (call/cc (#!lambda (k) k)))
    1000
    (#!lambda (x y) x)
    (#!lambda (x) x)))

(define-constant save
   (#!lambda (filename objectlist)
      (fluid-let ([output-port (open filename 'write)])
	  (pretty objectlist)
	  (close (fluid output-port))
	  t)))

(define-constant make-vector
   (lambda x (primitive-make-vector x)))

(define-constant vector (lambda x (list->vector x)))

(define-constant error
   (lambda args
      (newline)
      (print "Error: ")
      (mapc (lambda (x) (print x) (print `| |)) args)
      (reset)))

(define-constant unique (lambda () (ref 'unique)))
   
